#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <mpi.h>
#include <omp.h>
#include <signal.h>
#include <unistd.h>

// Esquema para el problema del juego de la vida con vecindad variable
// de la prueba online del concurso de programacion paralela 2011

void initialize(char *m,int t)
{
  int i;

// Se genera la matriz con valores 0 y 1, con la misma probabilidad
  for(i=0;i<t*t;i++)
  {  
    m[i] = (char) ((1.*rand()/RAND_MAX)*2.);
  }
}

void escribir(char *m, int t){
  int i, j;

  for (i = 0; i < t; i++)
  {
    for (j = 0; j < t; j++)
    {  
      printf("%d ",m[i*t+j]);
    }
    printf("\n");
  }
  printf("\n");
}

void escribirresult(char *m, int t,int cantidad){
  int i, j,cont=0,ceros=0,unos=0;

  for (i = 0; i < t; i++)
  {
    for (j = 0; j < t; j++)
    {  
      if(m[i*t+j]==0)
      {
        ceros++;
      }
      else
      {
        unos++;
      }
    }
  }
  printf("%d %d\n",ceros,unos);

  for (i = 0; i < t; i++)
  {
    for (j = 0; j < t; j++)
    {  
      if((cont%cantidad)==0)
      {
        printf("%d ",m[i*t+j]);
      }
      cont++;
    }
    printf("\n");
  }
  printf("\n");
}
/*
c
c     mseconds - returns elapsed milliseconds since Jan 1st, 1970.
c
*/
long long mseconds(){
  struct timeval t;
  gettimeofday(&t, NULL);
  return t.tv_sec*1000 + t.tv_usec/1000;
}

static void alarm_handler(int sig) {
  fprintf(stderr, "Time Limit Exceeded\n");
  abort();
}

void vida();

int main(int argc,char *argv[]) {
  int nodo,np,i,N,cuantos,semilla,vecmax,iteraciones,cantidadescribir;
  long long ti,tf,tt=0;
  char *a;

  FILE *stats_file = fopen("stats", "w");

  struct sigaction sact;
  sigemptyset(&sact.sa_mask);
  sact.sa_flags = 0;
  sact.sa_handler = alarm_handler;
  sigaction(SIGALRM, &sact, NULL);
  alarm(20);  /* time limit */
 
  MPI_Init(&argc,&argv);
  MPI_Comm_size(MPI_COMM_WORLD,&np);
  MPI_Comm_rank(MPI_COMM_WORLD,&nodo);

// Se lee y se distribuye a todos los procesos MPI
// el número de casos de prueba
  if(nodo==0)
  {
    scanf("%d",&cuantos);
    MPI_Bcast(&cuantos,1,MPI_INT,0,MPI_COMM_WORLD);
  }
  else
  {
    MPI_Bcast(&cuantos,1,MPI_INT,0,MPI_COMM_WORLD);
  }
  for(i=0;i<cuantos;i++)
  {
    if(nodo==0)
    {
      scanf("%d",&N);                   // El primer argumento es el tamano del tablero (cuadrado)
      scanf("%d",&semilla);             // El segundo una semilla para la generacion aleatoria de los datos
      scanf("%d",&vecmax);              // El tercero es el tamano máximo de la vecindad
      scanf("%d",&iteraciones);         // El cuarto el número de iteraciones a generar
      scanf("%d",&cantidadescribir);    // No se escribe el tablero completo, sino un número de valores separados un cierto espacio
                                        // se hace así para no guardar ficheros de resultado muy grandes
      a = (char *) calloc(sizeof(char),N*N);
      srand(semilla);
      initialize(a,N);
#ifdef DEBUG
// Mientras se está depurando se escribe el tablero
      escribir(a,N);
#endif
    }
    MPI_Barrier(MPI_COMM_WORLD);
    ti=mseconds(); 
    vida(N,a,vecmax,iteraciones,nodo,np);
    MPI_Barrier(MPI_COMM_WORLD);
    tf=mseconds(); 
    if(nodo==0)
    {
// El tiempo de la primera ejecución no se cuenta
      if(i!=0) tt+=tf-ti;
#ifdef DEBUG
// Mientras se depura se escribe el tiempo de cada ejecución
      fprintf(stats_file, "%Ld\n", tf-ti);
#endif
// Se escriben los resultados de todas las ejecuciones,
// incluso de la primera, de la que no se cuenta el tiempo
      escribirresult(a,N,cantidadescribir);
      free(a);
    } 
  }
  
  if(nodo==0)
  { 
    fprintf(stats_file, "%Ld\n", tt);
    fclose(stats_file);
  }
  MPI_Finalize();
  return 0;
}
